/*
Grouping Genetic Algorithm crossover operator.  
To be used only for grouping problems.
Last Modified 07/31/95                             
*/


#include "gga.h"


char which_crossover[] = "ggax.c";

void crossover(parent1, parent2, child1, child2, jcross)
  unsigned *parent1, *parent2, *child1, *child2;
  int jcross[2];
{
  struct group_chrom xover();


  int debug=0; /* set to 1 for debugging information */
  int count1;
  unsigned one=1; /* needed to convert int to unsigned */
  struct group_chrom parent1info, parent2info, child1info, child2info;
  int temp;

  jcross[0]=0; /* not used.  There will be 2 sites per chromosome. */
  jcross[1]=0; /* not used. */




  if(flip(pcross))
    {
      ncross++;

      parent1info=convert_from_binary(parent1);
      parent2info=convert_from_binary(parent2);

      parent1info.number_of_groups=determine_num_of_groups(parent1info);
      parent2info.number_of_groups=determine_num_of_groups(parent2info);
 
 
/*
Select at random two crossing sites,
delimiting the crossing section,
in each of the two parents.
*/
   parent1info.x1=rnd(0,parent1info.number_of_groups);
   parent1info.x2=rnd(0,parent1info.number_of_groups); 
   while(parent1info.x1==parent1info.x2)
     {
       parent1info.x2=rnd(0,parent1info.number_of_groups);
     }
   if(parent1info.x1>parent1info.x2)
     {
       temp=parent1info.x1;
       parent1info.x1=parent1info.x2;
       parent1info.x2=temp;
     }
     
   parent2info.x1=rnd(0,parent2info.number_of_groups);
   parent2info.x2=rnd(0,parent2info.number_of_groups);
   while(parent2info.x1==parent2info.x2)
     {
       parent2info.x2=rnd(0,parent2info.number_of_groups);
     }
   if(parent2info.x1>parent2info.x2)
     {
       temp=parent2info.x1;
       parent2info.x1=parent2info.x2;
       parent2info.x2=temp;
     }


/*
Child 1 is created by injecting the groups from parent2, keeping
the unaffected groups from parent1, and repacking the remaining objects.
Child 2 is created with the roles of the parents reversed.
*/
   child1info=xover(parent1info,parent2info);
   child2info=xover(parent2info,parent1info);
  

								     
if(debug==1)
  {
    fprintf(outfp,"\n\nggax.c");
    fprintf(outfp,"\nParent1");
    print_chr(parent1info);
    fprintf(outfp,"\nParent2");
    print_chr(parent2info);
    fprintf(outfp,"\nChild1");
    print_chr(child1info);
    fprintf(outfp,"\nChild2");
    print_chr(child2info);
  }


  make_binary(child1info,child1);
  make_binary(child2info,child2);   
  
} /* if flip(pcross) */  
else
  {
    for(count1=0; count1<chromsize; count1++)
      {
	child1[count1]=parent1[count1];
	child2[count1]=parent2[count1];
      }
   }

}





struct group_chrom xover(parent1info, parent2info)
struct group_chrom parent1info, parent2info;
{  
  int debug=0; /* set to 1 for debugging information */
  int groups_affected[MAX_NUM_OF_OBJECTS];
  int objects_moving[MAX_NUM_OF_OBJECTS];
  int in_queue[MAX_NUM_OF_OBJECTS];
  int num_in_queue;
  int count1, count2;
  struct group_chrom temp1;
  int queue[MAX_NUM_OF_OBJECTS]; 
  struct group_chrom remove_smallest_group();
/*
Inject the contents of the crossing section of the first parent
at the first crossing site of the second parent.  Recall that 
the crossover works with the group part of the chromosome, so this 
means injecting some of the groups from the first parent into the 
second.

child1=parent1 with parent2 injected.
child2=parent2 with parent1 injected.
*/
     
/*
Membership in the injected groups has precedence over old groups.
Determine all objects moving to new groups injected from parent2.
*/
  for(count1=0; count1<number_of_objects; count1++)
    {
      objects_moving[count1]=0;
      groups_affected[count1]=0;
      in_queue[count1]=0;
    }
  for(count1=parent2info.x1; count1<parent2info.x2; count1++)
    {
      for(count2=0; count2<number_of_objects; count2++)
	{
	  if(parent2info.object[count2]==parent2info.group[count1])
	    {
	      objects_moving[count2]=1;
	      groups_affected[parent1info.object[count2]]=1;
	    }
	}      
    }


/*
If one or more object moves out of a group, the group is deleted from
the solution.  Find all groups that are affected by at least one object
moving.  All objects remaining in these affected groups are put 
in the queue to be later reinserted back into the solution.
*/
 for(count1=0; count1<number_of_objects; count1++) 
   {
     if((groups_affected[parent1info.object[count1]]==1)&&(objects_moving[count1]==0))
       {
	 in_queue[count1]=1;
       }
   }



/*
Create group part of temp1
Groups from parent2 are injected into parent 1.
Groups from parent1 are kept ONLY IF they contain no objects
that belong to the injected groups.
*/
  count1=0;
  for(count2=0; count2<parent1info.x1; count2++)
    {
      if(groups_affected[parent1info.group[count2]]==0)
	{
	  temp1.group[count1]=parent1info.group[count2];
	  count1++;
	}
    }
  for(count2=parent2info.x1; count2<parent2info.x2; count2++)
    {
      temp1.group[count1]=parent2info.group[count2]+number_of_objects;
      count1++;
    }
  for(count2=parent1info.x1; count2<parent1info.number_of_groups; count2++)
    {
      if(groups_affected[parent1info.group[count2]]==0)
	{
	  temp1.group[count1]=parent1info.group[count2];
	  count1++;
	}
    }
  temp1.number_of_groups=count1;
if(debug==1)
  {
    fprintf(outfp,"\n\nCreate group part of temp1");
    print_chr(temp1);
  }  
 


/*
If the object is moving to an injected group (objects_moving[count1]==1),
then put it in the injected group.  Note that (number_of_objects) is
added to the injected group name to guarantee that no name conflict will
occur with any group name in parent1.  The groups will all be renamed
later.
If the object is not moving and is not in the queue (in_queue[count1]
==0), then the group retains its membership in the group from parent1.
If the object is in the queue, it will be reinserted later. 
*/
  for(count1=0; count1<number_of_objects; count1++)
    {
      if(objects_moving[count1]==1)
	{
	  temp1.object[count1]=parent2info.object[count1]+number_of_objects;
	}
      else if(in_queue[count1]==0)
	{
	  temp1.object[count1]=parent1info.object[count1];
	}
    }
if(debug==1)
  {
    fprintf(outfp,"\n\nCreate object part of temp1");
    print_chr(temp1);
  }


/* 
Fill the queue with objects in the queue. Keep a count 
of the number of delete objects 
*/
  num_in_queue=0;
  for(count1=0; count1<number_of_objects; count1++)
    {
       if(in_queue[count1]==1)
	 {
	   queue[num_in_queue]=count1;
	   num_in_queue++;
	 }
    }




return application_dependent_reinsert(temp1,queue, num_in_queue,in_queue);


}






	     
