/*----------------------------------------------------------------------------*/
/* sga.h - global declarations for main(), all variable declared herein must  */
/*         also be defined as extern variables in external.h !!!              */
/*----------------------------------------------------------------------------*/

/* If your C compiler accepts the newer ANSI-style function definitions    */
/* with prototypes included, such as:  int dummyfunc(int arg1, float arg2) */
/* then you should leave the next line defined as is.  If not, simply      */
/* comment out the next line.                                              */

/* #define PROTOTYPES_ACCEPTED */
/* WARNING!  These are also defined in external.h!                         */
/* See also discussion of SECONDARY_PROTOTYPES_ACCEPTED in header for      */
/* file external.h, and comment out next line if needed for your compiler  */
/* #define SECONDARY_PROTOTYPES_ACCEPTED */

#include <stdlib.h>
#include <stdio.h>


#ifndef _SGA_DEFINE

#define _SGA_DEFINE


#define LINELENGTH 70   /* width of printout */
#define BITS_PER_BYTE 8 /* number of bits per byte on this machine */
#define UINTSIZE (BITS_PER_BYTE*sizeof(unsigned))  /* # of bits in unsigned */
#define MAXSCALINGWINDOW 20

/* The next 3 defines may be enlarged freely if need more fields (in a        */
/* reordering-type or mixed-type problem), or more variables in a mixed-type  */
/* representation.. limited only by the memory capacity of your machine.      */
#define MAXNUMFIELDS 100

typedef unsigned int BOOL;

/* file pointers */
FILE   *outfp, *infp, *plotfp;

/* Global structures and variables */
struct individual {
    unsigned *chrom;    /* chromosome string for the individual */
    double  fitness;  /* fitness of the individual, after (optional) scaling */
    double  init_fitness; /* raw (unscaled) fitness of individual */
    long    neval;      /* number of evaluation of the Royal Road Function */
    int     xsite[2];   /* crossover site(s) at mating */
    int     parent[2];  /* who the parents of offspring were */
    int    *utility;    /* utility field can be used as pointer to a */
	    /* dynamically allocated, application-specific data structure */
};

struct bestever {
    unsigned *chrom;    /* chromosome string for the best-ever individual */
    double  fitness;    /* scaled fitness of best-ever individual of popn */
    double  init_fitness; /* unscaled fitness of best-ever individual of popn */
    long    neval;      /* number of evaluation of the Objective Function */
    int     generation; /* generation which produced it */
    int    *map_to_std; /* inversion map_to_std needed to transform this */
			/* individual's chromosome to 'standard' rep used */
			/* in the objective function.  (identity map when */
			/* not using_inversion. */
    int    *field_ends; /* also need ends of fields, for decoding best */
			/* individuals when their inversion pattern no */
			/* longer matches the current one. */
    int    *utility;  /* pointer to user-defined utility fields */
};

struct individual *oldpop;      /* last generation of individuals */

struct individual *newpop;      /* next generation of individuals */

struct individual kidholder[2]; /* temporary holder for offspring */

struct bestever bestfit;        /* fittest individual so far, for this pop    */
				/* (if ESGA) or all subpops THIS processor    */
				/* (if IPGA) */

struct bestever local_bestfit;     /* for parallel subpops, best among all */
				/* populations which THIS process (or */
				/* processor) calculates */

struct bestever all_pops_bestfit; /* fittest individual so far, ALL pops,     */
				  /* ALL processors using this file directory */

unsigned *mapped_saved_chromosome;    /* pointer to chromosome in the "native" form */
				/* in terms of which the objective function   */
				/* objfunc is defined.                        */

int    *this_pop_map_to_std;    /* ptr to map showing, for each field on this */
		       /* subpop's chromosomes, which field that is on the    */
		       /* "unmapped" form delivered to the objective function */

int    *donor_pop_map_to_std;  /* ptr to map showing, for each field on donor */
		       /* subpop's chromosomes, which field that is on the    */
		       /* "unmapped" form delivered to the objective function */

int    *donor_pop_field_ends;  /* ptr to array indicating bit posn of end of  */
		       /* each field on chromosome (may be different lengths, */
		       /* and fields may be reshuffled by inversion,          */
		       /* different reps, etc. */

double  sumfitness;             /* summed fitness for entire population, THIS */
				/* generation only */

double one_pop_cum_sumfitness=0.;   /* summed fitness for entire population */
				    /* to date, only this subpopulation */

double one_pop_sum_best_fitness=0.; /* sum of best fitness of each generation */
				    /* for THIS subpopulation only. */

long    one_pop_best_fitness_count = 0; /* count number of times sumfitness */
				 /* has been added to one_pop_cum_sumfitness, */
				 /* which is number of best fitnesses calc'd */

long    one_pop_online_denominator = 0; /* track total indiv's in THIS pop */
					/* for all gens, even if popsize is */
					/* changed.  */

double  all_pops_sumfitness=0.;  /* For Manypops, summed fitness for all */
				 /* individuals evaluated to date */

double  all_pops_sum_best_fitness = 0.;   /* For Manypops, sum of best */
		     /* fitness of ALL generations across all populations */

long    all_pops_online_denominator = 0; /* For Manypops, denominator for */
				 /* use in calculating on-line performance */
				 /* across all subpopulations */

long    all_pops_best_fitness_count = 0;  /* For Manypops, count of number */
				 /* of times sumfitness is evaluated across */ 
				 /* ALL subpopulations to date, which is */
				 /* same as number of best fitness calc'ns */

long    all_pops_neval = 0;/* For Manypops, TOTAL number of evaluations */
				 /* performed to date, over ALL */
				 /* subpopulations */
				 
double  local_cycle_sumfitness = 0.; /* For Manypops, tracks stats within a */
				     /* cycle.  Needed for updating all_pops */
				     /* stats at end of a cycle, keeping     */
				     /* good stats even across MULTIPLE      */
				     /* processors. */

double  local_cycle_sum_best_fitness = 0.; /* as above */

long    local_cycle_online_denominator = 0; /* as above */

long    local_cycle_best_fitness_count = 0; /* as above */

long    local_cycle_neval = 0; /* as above */

int     new_bestfit_neval_within_local_cycle = 0;

double  bestfit_init_fitness_within_local_cycle = 0.;

int     popno;  /* Number [0, npops-1] of current subpopulation; 0 for Onepop */

int     npops;  /* Number of subpopulations to simulate; 1 for Onepop */

int     cycle;  /* Number of current cycle of subpopulations handled by */
		/* this process or processor;  0 for Onepop */ 

int     ncycles; /* Number of cycles to run before halting (Manypops); */
		 /* 0 for Onepop */

int     startpopnum;       /* see immediately below */

int     finishpopnum;      /* for TRULY parallel (multiple CPU's) runs */
      /* of Manypops, or if subpopulation sets are run from MORE THAN ONE */
      /* process on the same processor, tells index (in master file) of */
      /* the starting and stopping subpopulations to be calculated by THIS */
      /* process or processor.    */

BOOL    single_process; /* FALSE means more than one processor (or process) */
			  /* is running subpopulations on the same problem */
			  /* in the same directory */

BOOL    bestever_utilities_mallocd = 0; /* Flag is TRUE (==1) after malloc */
		  /* utility fields in the bestever structures declared. */

BOOL    bestever_chroms_mallocd = 0; /* Flag is TRUE (==1) after malloc */
		 /* chrom fields in the bestever structures declared. */

/* Next is flag to determine whether new parameters should be read in for */
/* each subpopulation (Manypops only), or the first set read should be used */
/* for all of them (from startpopnum to finishpopnum). */
BOOL    all_subpops_use_same_parameters = 1; /* starts TRUE */

double  minraw; /* minimum raw fitness in population now */

double  avgraw; /* average raw fitness in population now */

double  fit_max;        /* maximum fitness of population (after any scaling) */

double  fit_avg;        /* average fitness of population (after any scaling) */

double  fit_min;        /* minumum fitness of population (after any scaling) */

int     permproblem;    /* 1 indicates permutation problem; 0, otherwise. */

int     alpha_size;     /* Number of different valid bit patterns per field */
			/* If set <2, means different-sized fields, use sizes */
			/* from array field_sizes in app_set_options. */

int    *field_sizes;    /* array of ints used to specify alphabet size for */
			/* each field when they differ -- used only when */
			/* indicated by user setting alpha_size < 2 */

int    *field_ends;     /* number (in [1,lchrom]) of final bit of ith field. */
			/* Fields are numbered [0, numfields-1]. */

int    *std_field_ends; /* Ends of fields for STANDARD (initial) pattern, */
			/* also used in fitness function.  Never changes. */

int     numfields;      /* number of fields on the chromosome (= length   */
			/* in bits, if alpha_size = 2)                    */

int     max_numfields_among_all_subpops; /* User sets to largest */
					     /* value for numfields used in */
					     /* ANY subpop (== numfields, */
					     /* unless different_reps == 1) */

int  incest_reduction; /* >1 if want to use Goodman's method  */
		/* for favoring crossovers among fit individuals which */
		/* are relatively distant in the Hamming sense.        */

BOOL user_supplied_initialization;         /* Must set TRUE in app_init() */
					   /* if user will supply own code */
					   /* for initializing individuals */

BOOL call_app_user_does_migration;         /* Flag, if set, allows user */
					   /* full control of migration */

BOOL different_reps;        /* Flag, if set, says different subpops  */
			    /* use different representations.  Means */
			    /* that USER MUST supply code in routine */
			    /* app_transform_and_move_migrant() in   */
			    /* the user's application file.          */

BOOL    using_inversion;    /* Flag, if set, says that each subpop   */
			    /* in Manypops has its own inversion     */
			    /* pattern (mapping of the chromosome),  */
			    /* and will periodically be remapped to  */
			    /* another pattern.  Implies that when   */
			    /* objective function is called, the     */
			    /* chromosome must first be "uninverted" */
			    /* and when migration occurs, must unmap */
			    /* chromosome, then remap to new subpop's*/
			    /* inversion pattern.                    */
int     parameter_changes;
int     ckptfreq;

int     fieldlength;    /* length (in bits) of each field (1 for binary) */

int     crosstype;      /* holds code for type of crossover operator desired */

int     mutatetype;     /* holds code for type of mutation operator desired */

float   pcross = -1.;         /* probability of crossover; illegal default */
			/* will be replaced in user_in */

float   pmutation = -1.; /* probability of mutation; illegal default */
			/* will be replaced in user_in. */

float   p_multimut = -1.; /* probability of performing a specialized, */
			  /* multiple-consecutive-locus mutation. */

float   avg_mut_width = 1.; /* Expected value of number of loci altered by */
			    /* the multi-bit mutation operator when it does */
			    /* a mutation. */

float   pinversion;     /* probability of inversion (yielding a new map */
			/* for entire subpopulation).  Expressed as     */
			/* probability PER GENERATION.                  */

int     autocontrol_selection = 1; /* Boltzmann selection only; 1 means */
			/* controlling automatically each generation */
			/* using beta_s_tilde; otherwise, use user-supplied */
			/* beta, unchanged. */
double  beta = -1.;    /* for Boltzmann selection only, default value "off" */

float   beta_s_tilde; /* User-suppliable value defaults to:        */
		      /* 0.3*sqrt(2.*log(popsize)) */
		      /* unless user changes.  Beta calculated from it by */
		      /* dividing it by std dev. of fitness distrib. each */
		      /* generation, when autocontrol_selection is true xx*/
		      /* and Boltzmann selection is active. */

float   scalemult;      /* mult of avg fitness to which most fit indiv is */
			/* scaled, so long as produces no neg values. */
			/* Used for same purpose when rank-based selection */
			/* is used. */

float   conv_measure;   /* fraction of hi-fitness guys NOT closest */
			/* (Hamming) to best current individual    */

float   percent_lost;   /* percent of loci with all 0's or all 1's */

float   percent_converged; /* percent of loci with >=95% 0's or >= 95% 1's */

float   avg_bias; /* average bias (percent same allele) of all loci */


double  on_line_performance; /* Average of ALL fitnesses of ALL individuals */
			     /* evaluated over entire run (for one */
			     /* subpopulation, if Manypops) */

double  off_line_performance; /* Average of BEST individual's fitness for ALL */
			      /* generations in entire run (for one */
			      /* subpopulation, if Manypops) */

double  all_pops_on_line_performance; /* Average on_line_performance across */
			      /* ALL subpopulations (should not be  */
			      /* used from Onepop, only from Manypops) */

double all_pops_off_line_performance; /* Average off_line_performance across */
			      /* ALL subpopulations (should not be   */
			      /* used from Onepop, only from Manypops)     */

float   conv_sigma_coeff;       /* mult of sigma; guys with fitness above  */
				/* this no. of std dev'ns above mean are   */
				/* examined for sim to best guy, using     */
				/* Hamming distance.                       */

float   sigma_trunc;    /* multiplier for sigma, per Goldberg, p.125 */

float   sigma;          /* std deviation of current population */

long    neval;  /* number of evaluations of the objective function, this pop */

int     numfiles;       /* number of files specified on cmd line */
int     interactive;    /* input state */
int     extending;        /* starting a continuation run from previous state */
int     popsize;        /* population size */

int     lchrom; /* length of the chromosome per individual */

int     chromsize; /* number of unsigned ints needed to store lchrom * string */

int     max_chromsize_among_all_subpops; /* User sets to largest chromsize */
					 /* used in ANY subpop (== chromsize, */
					 /* unless different_reps == 1) */

int     genspercycle = 10;      /* num. of generations to run island parallel */
				/* GA (Manypops) before checkpointing */

int     startgen;       /* number of starting generation, !=0 for restart   */

int     gen;    /* current generation number */

int     maxgen; /* maximum generation number */

int     run;    /* current run number */

int     maxruns;        /* maximum number of runs to make */

int     tourneysize = 2;  /* for size of tournament if tselect.c is in use  */
			  /* (default = 2).  Otherwise, ignored.) */

int     printstrings = 0;       /* flag to print chromosome strings (default
				 * off) */

long    nmutation;      /* number of mutations */

long    ncross;         /* number of crossovers */

int     bestnow;        /* index of best individual of current generation */

int    *ones;   /* Array counting 1's at each locus, to look for      */
		/* misaligns, convergence, etc. */

int     convinterval = 10;   /* generations between reporting convergence */
			/* statistics (number of ones per locus) */

int     crowding_factor;  /* for DeJong-style crowding; 0 means none   */

int     stochastic;     /* if 1, all chroms. eval'd every generation; if
			 * 0,not. */

int     elitism;        /* if 1, best of gen. always present in next gen.; if
			 * 0, not. */

int     quiet = 0;  /* suppresses most output except from app-supplied routines */

int     scaling_window; /* size of scaling window (-1 = no window scaling) */

int     windowstart;    /* pointer for window scaling */

int     windowend;      /* pointer for window scaling */

double  savedmins[MAXSCALINGWINDOW];    /* min fitness history, for   */
					/* window scaling */

int    *available;    /* dynam. alloc. in memory.c */

int    *evaluate;       /* malloc popsize flags saying need to call objfunc. */

/* Below are some variables used by various permutation-type operators, for   */
/* which space is malloc'ed only if permutation-type problem being run.       */

/* CAUTION:  If you load permutation-type operators, you MUST answer 'y' to   */
/* the question "Permutation-type problem?"                                   */
int    *intarray1 = NULL;

int    *intarray2 = NULL;

int    *intarraybak1 = NULL;

int    *intarraybak2 = NULL;

int    *randomlst = NULL;

int    *tmparray = NULL;


/* File name arrays (make prefix arrays WAY too long, in case of bad input */
char    restartfileprefix[29], restarthdrfilename[13], restartindfilename[13];

char    masterfileprefix[29], masterfilename[13];

char    checkpthdrfilename[13], checkptindfilename[13], checkptfileprefix[29];


/* variables for random number generator, so can preserve them for restart */
double  oldrand[55];

int     jrand;

double  rndx2;

int     rndcalcflag;

float   randomseed;

/* For optional superuniform initialization */
int superuniform_cyclebitwidth = 0;
int superuniform_initialization = 0; /* FALSE not defined, yet. */

int many;

/* Application-dependent declarations go after here... */

#endif


#ifdef PROTOTYPES_ACCEPTED
#include "sgafunc.h"

#else
#include "sgapure.h"

#endif


