/*----------------------------------------------------------------------------*/
/* checkwt.c -- Main checkpoint routine, write_checkpoint, plus routines to   */
/*              write the checkpoint files containing the individuals         */
/*              (chromosomes, etc.) in the population.  Includes FileRename,  */
/*              a routine used to change the name of the individual files     */
/*              with suffix .new to suffix .ind and with suffix .neu to .ckp. */
/*              NOTE:  If the user modifies the struct individual, user must  */
/*              also modify individualsize in this file.                      */
/*----------------------------------------------------------------------------*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "external.h"

#define RETRIES 10

static FILE *fpwrite = NULL;
static int Version = 1;
static int IndCnt;
static int IndSize;
static int ChromSize;
static long states[2];

#ifdef PROTOTYPES_ACCEPTED
static BOOL InitCPWrite(char *fname, int icnt, int isize, int csize);
static BOOL EndCPWrite(char *fname);
static BOOL AddInd(struct individual*, BOOL(*)());
#else
static BOOL InitCPWrite();
static BOOL EndCPWrite();
static BOOL AddInd();
#endif

void write_checkpoint()
{
  int     j;
  int     individualsize;

  /* *** NOTICE: If you modify what is written to the .ind files for */
  /* an individual, you must also fix individualsize below.            */
  individualsize = 2 * sizeof(double) + 4 * sizeof(int) + sizeof(long)
    + chromsize * sizeof(unsigned int) + GetUtilitySize();
  
  /* Start to write the .ind file (need not be locked) */
  if (!cycle && !startgen)
    init_write(checkptindfilename);
  InitCPWrite(checkptindfilename, popsize, individualsize, chromsize);
  
  for (j = 0; j < popsize; j++)
    if (AddInd(&(oldpop[j]), sys_util_write_utility) == FALSE) {
      fprintf(stderr, "\nError:Write ind error at indiv # %d.\n", j);
      break;
    }

  EndCPWrite(checkptindfilename);
  if (!cycle && !startgen)
    init_write(checkpthdrfilename);
  WriteCheckPointHeader(checkpthdrfilename, app_write_ckp_hdr);
  return;
}


static BOOL
InitCPWrite(fname, icnt, isize, csize)
char   *fname;
int     icnt, isize, csize;
{
  IndCnt = icnt;
  IndSize = isize;
  ChromSize = csize;
  
  if (!(fpwrite = start_write(fname, states, RETRIES))) {
    fprintf(stderr, "\n*** ERROR: Unable to open %s for write.\n",fname);
    return FALSE;
  }
  
  if (fwrite(&Version, sizeof(int), 1, fpwrite) == 1)
    if (fwrite(&icnt, sizeof(int), 1, fpwrite) == 1)
      if (fwrite(&isize, sizeof(int), 1, fpwrite) == 1)
	if (fwrite(&csize, sizeof(int), 1, fpwrite) == 1)
	  if (fwrite(&bestnow, sizeof(int), 1, fpwrite) == 1)
	    return TRUE;
  
  return FALSE;
}

static BOOL
EndCPWrite(fname)
char *fname;
{
  if (end_write(fname, fpwrite, states, RETRIES)) {
    fpwrite = NULL;
    return TRUE;
  }
  
  fpwrite = NULL;
  return FALSE;
}

static BOOL
AddInd(IndBuf, CBWrite)
struct individual *IndBuf;
BOOL(*CBWrite) ( /* int *,FILE * */ );
{
  if (!fpwrite)
    return FALSE;

  if (fwrite((char *)IndBuf->chrom, sizeof(unsigned int), ChromSize, fpwrite)
      == ChromSize)
    if (fwrite(&(IndBuf->fitness), sizeof(double), 1, fpwrite) == 1)
      if (fwrite(&(IndBuf->init_fitness), sizeof(double), 1, fpwrite) == 1)
	if (fwrite(&(IndBuf->neval), sizeof(long), 1, fpwrite) == 1)
	  if (fwrite(&(IndBuf->xsite[0]), sizeof(int), 1, fpwrite) == 1)
	    if (fwrite(&(IndBuf->xsite[1]), sizeof(int), 1, fpwrite) == 1)
	      if (fwrite(&(IndBuf->parent[0]), sizeof(int), 1, fpwrite) == 1)
		if (fwrite(&(IndBuf->parent[1]), sizeof(int), 1, fpwrite) == 1)
		  return (*CBWrite) (IndBuf->utility, fpwrite);

  return FALSE;
}

BOOL
sys_util_write_utility(UtilityBuf, fp)
int    *UtilityBuf;
FILE   *fp;

/* Application-dependent callback routine which provides for writing the  */
/* contents of the utility field into a checkpoint file, IF THE USER DEFINES */
/* utility fields (GetUtilitySize() > 0).                */
{
  unsigned nbytes;

  nbytes = GetUtilitySize();
  if (nbytes)
    if (fwrite(UtilityBuf, nbytes, 1, fp) != 1)
	return FALSE;
  return TRUE;
}



