/*----------------------------------------------------------------------------*/
/* oneptx.c -- Single-point crossover for ordinary (non-permutation) problems */
/*             Crossover done with probability pcross; else, parents copied   */
/*             unaltered to children.  Works for binary chromosomes or those  */
/*             with alphabet cardinality >2, but NOT for permutation-type.    */
/*             Modified 4/96 by EDG to work with different length fields,     */
/*             and still to do crosses only at field boundaries for           */
/*             non-binary chromsomes, whether field lengths same or not.      */
/*----------------------------------------------------------------------------*/

#include "external.h"

char which_crossover[] = "oneptx.c";

void
crossover(parent1, parent2, child1, child2, jcross)
unsigned *parent1, *parent2, *child1, *child2;
int     jcross[2];

/* Cross 2 parent strings, place in 2 child strings */
{
    int     j, k, jcross1;
    unsigned mask, temp;

    /* Do crossover with probability pcross */
    if (flip(pcross)) {
	/* Calculate after which bit to break string, based on fields. */
	/* Note that cannot break after field_ends[numfields-1], as that */
	/* is at the end of the chromosome, so produces no crossover. */
	jcross1 = field_ends[rnd(0, numfields - 2)];
	jcross[0] = jcross1;
	jcross[1] = 0;	/* Not used in one-point crossover */
	ncross++;
	for (k = 1; k <= chromsize; k++) {
	    if (jcross1 >= (k * UINTSIZE)) {
		child1[k - 1] = parent1[k - 1];
		child2[k - 1] = parent2[k - 1];
	    }
	    else if ((jcross1 < (k * UINTSIZE)) && (jcross1 > ((k - 1) * UINTSIZE))) {
		mask = 1;
		for (j = 1; j <= (jcross1 - 1 - ((k - 1) * UINTSIZE)); j++) {
		    temp = 1;
		    mask = mask << 1;
		    mask = mask | temp;
		}
		child1[k - 1] = (parent1[k - 1] & mask) | (parent2[k - 1] & (~mask));
		child2[k - 1] = (parent1[k - 1] & (~mask)) | (parent2[k - 1] & mask);
	    }
	    else {
		child1[k - 1] = parent2[k - 1];
		child2[k - 1] = parent1[k - 1];
	    }
	}
    }
    else {
	for (k = 0; k < chromsize; k++) {
	    child1[k] = parent1[k];
	    child2[k] = parent2[k];
	}
	jcross[0] = 0;
	jcross[1] = 0;
    }
}
