/*----------------------------------------------------------------------------*/
/* scramble.c -- Random Sublist Scramble Mutation -- a mutation operator for  */
/*               use on permutation-type problems.  As described in Davis,    */
/*               Handbook of Genetic Algorithms.  Does a random shuffle of    */
/*               the fields between two randomly selected points on the       */
/*               chromosome.                                                  */
/*----------------------------------------------------------------------------*/

#include "external.h"
#include <stdlib.h>

char which_mutation[] = "scramble.c";

void
mutation(child)
unsigned *child;

/* Perform Random Sublist Scramble Mutation (as suggested by Davis, Handbook  */
/* of Genetic Algorithms), on an individual,  with probability pmutation, and */
/* count # of mutations.  Note... in this case, pmutation is interpreted as   */
/* probability PER INDIVIDUAL, NOT the usual probability PER BIT.             */
/* NOTE:  If you use this mutation operator, you must also select ONE of the  */
/* permutation-type crossover operators (their files contain FirstNth(), which*/
/* is called here).                                                           */
{
    int     i, j, k;
    int     cnt, start, end;

    if (!flip(pmutation))
	return;
    nmutation++;

    j = rnd(0, numfields - 1);
    k = rnd(0, numfields - 1);
    start = (j < k) ? j : k;
    end = (j > k) ? j : k;
    cnt = end - start + 1;

    for (i = 0; i < numfields; i++)
	tmparray[i] = (i < start || i > end) ? 0 : 1;

    chromtointarray(intarray1, child);
    chromtointarray(intarraybak1, child);

    for (i = start; i <= end; i++) {
	k = rnd(0, cnt - 1);
	j = FirstNth(tmparray, k, start, end);
	tmparray[j] = 0;
	intarray1[i] = intarraybak1[j];
    }

    intarraytochrom(intarray1, child);
}


void
field_mutation(child)
unsigned *child;
/* DUMMY FUNCTION for compatibility with non-perm.-type mutation operators. */
{
}

