/*----------------------------------------------------------------------------*/
/* twoptx.c -- Two-point (or circular) crossover operator for ordinary        */
/*             (binary, non-permutation) problems:  treats chromosome as a    */
/*             ring, with probability pcross, child1 gets part from locus     */
/*             jcross1 to jcross2 from parent 2, rest from parent 1; child 2  */
/*             is reverse; else, children are copied unaltered from parents.  */
/*             Works for binary chromosomes or those with alphabet cardinality*/
/*             >2, but not for permutation-type problems.                     */
/*             Modified 4/96 by EDG to cross at field boundaries even if use  */
/*             different length fields.  (probab. is indep of fld lengths)    */
/*----------------------------------------------------------------------------*/

#include "external.h"

char which_crossover[] = "twoptx.c";

void
crossover(parent1, parent2, child1, child2, jcross)
unsigned *parent1, *parent2, *child1, *child2;
int     jcross[2];

/* Cross 2 parent strings, place in 2 child strings */
{
    int     j, k, jcross1, jcross2;
    unsigned mask, temp;

    /* Do crossover with probability pcross */
    if (flip(pcross)) {
	jcross1 = field_ends[rnd(0, numfields - 1)];   /* Cross at any field */
	/* boundary */
	jcross2 = field_ends[rnd(0, numfields - 1)];	/* same */
	ncross++;
	if (jcross1 > jcross2) {
	    j = jcross2;
	    jcross2 = jcross1;
	    jcross1 = j;
	}
	jcross[0] = jcross1;
	jcross[1] = jcross2;
	for (k = 1; k <= chromsize; k++) {
	    if (jcross1 >= (k * UINTSIZE)) {
		child1[k - 1] = parent1[k - 1];
		child2[k - 1] = parent2[k - 1];
	    }

	    else if ((jcross1 < (k * UINTSIZE)) && (jcross1 > ((k - 1) * UINTSIZE))) {
		mask = 1;
		for (j = 1; j <= (jcross1 - 1 - ((k - 1) * UINTSIZE)); j++) {
		    temp = 1;
		    mask = mask << 1;
		    mask = mask | temp;
		}
		child1[k - 1] = (parent1[k - 1] & mask) | (parent2[k - 1] & (~mask));
		child2[k - 1] = (parent1[k - 1] & (~mask)) | (parent2[k - 1] & mask);
		if (jcross2 < (k * UINTSIZE)) {
		    /* handle case when both in same unsigned int, so already */
		    /* did some work into child1 and child2.                  */
		    mask = 1;
		    for (j = 1; j <= (jcross2 - 1 - ((k - 1) * UINTSIZE)); j++) {
			temp = 1;
			mask = mask << 1;
			mask = mask | temp;
		    }
		    child1[k - 1] = (child1[k - 1] & mask) | (parent1[k - 1] & (~mask));
		    child2[k - 1] = (parent2[k - 1] & (~mask)) | (child2[k - 1] & mask);
		}
	    }

	    else if (jcross2 >= (k * UINTSIZE)) {
		child1[k - 1] = parent2[k - 1];
		child2[k - 1] = parent1[k - 1];
	    }

	    else if ((jcross2 < (k * UINTSIZE)) && (jcross2 > ((k - 1) * UINTSIZE))) {
		mask = 1;
		for (j = 1; j <= (jcross2 - 1 - ((k - 1) * UINTSIZE)); j++) {
		    temp = 1;
		    mask = mask << 1;
		    mask = mask | temp;
		}
		child2[k - 1] = (parent1[k - 1] & mask) | (parent2[k - 1] & (~mask));
		child1[k - 1] = (parent1[k - 1] & (~mask)) | (parent2[k - 1] & mask);
	    }
	    else {
		child2[k - 1] = parent2[k - 1];
		child1[k - 1] = parent1[k - 1];
	    }

	}
    }
    else {
	for (k = 0; k < chromsize; k++) {
	    child1[k] = parent1[k];
	    child2[k] = parent2[k];
	}
	jcross[0] = 0;
	jcross[1] = 0;
    }
}
