/*----------------------------------------------------------------------------*/
/* uobx.c -- Uniform Order-Based Crossover (see Davis, Handbook of Genetic    */
/*                 Algorithms) for use in solving permutation-type            */
/*           problems.  Also includes FirstNth(), a function called by        */
/*           mutation operators for permutation-type problems, to assure that */
/*           compatible crossover and mutation operators are loaded. */
/*----------------------------------------------------------------------------*/

#include "external.h"
#include <stdlib.h>

char which_crossover[] = "uobx.c";

void
crossover(parent1, parent2, child1, child2, jcross)
unsigned *parent1, *parent2, *child1, *child2;
int     jcross[2];

/* Cross 2 parent strings, place in 2 child strings */
{
    int     i, j, k, cnt, realcnt;

    /* Do crossover with probability pcross */
    if (flip(pcross)) {
	/* Notice : this function returns jcross entries as 0, for the */
	/* position of crossover (jcross not useful). */
	ncross++;
	chromtointarray(intarray1, parent1);
	chromtointarray(intarraybak1, parent1);
	chromtointarray(intarray2, parent2);
	chromtointarray(intarraybak2, parent2);
	for (i = 0; i < numfields; i++)
	    randomlst[i] = rnd(0, 1);
	/* here generate child1. */
	cnt = 0;
	realcnt = 0;
	for (i = 0; i < numfields; i++)
	    if (randomlst[i] == 0)
		tmparray[cnt++] = i;
	for (i = 0; i < numfields; i++)
	    for (j = 0; j < cnt; j++)
		if (intarraybak2[i] == intarraybak1[tmparray[j]])
		    intarray1[tmparray[realcnt++]] = intarraybak1[tmparray[j]];
	/* here generate child2. */
	cnt = 0;
	realcnt = 0;
	for (i = 0; i < numfields; i++)
	    if (randomlst[i] == 1)  {
		tmparray[cnt++] = i;
	    }
	for (i = 0; i < numfields; i++)
	    for (j = 0; j < cnt; j++)
		if (intarraybak1[i] == intarraybak2[tmparray[j]]) {
		    intarray2[tmparray[realcnt++]] = intarraybak2[tmparray[j]];
		}
	intarraytochrom(intarray1, child1);
	intarraytochrom(intarray2, child2);
    }
    else {
	for (k = 0; k < chromsize; k++) {
	    child1[k] = parent1[k];
	    child2[k] = parent2[k];
	}
    }
    jcross[0] = 0;
    jcross[1] = 0;
}


int
FirstNth(array, k, start, end)
int    *array;
int     k, start, end;
{
    int     cnt, i;

    cnt = 0;
    for (i = start; i <= end; i++)
	if (array[i])
	    cnt++;
    if (!cnt) {
	fprintf(stderr, "\nTemplate is wrong in FirstNth.\n");
	exit(-1);
    }
    cnt = k % cnt;
    for (i = start; i <= end; i++) {
	if (array[i]) {
	    if (!cnt)
		return i;
	    else
		cnt--;
	}
    }
    fprintf(stderr, "\nTemplate wrong in FirstNth.\n");
    exit(-1);
}
