/*----------------------------------------------------------------------------*/
/* utility.c - utility routines; contains copyright, repchar, skip,       */
/*             ithruj2int, int2ithruj, putfield, getfield, powerof2,      */
/*             ceilinglogbase2, floorlogbase2, chromtointarray,           */
/*             intarraytochrom, hamming, indextofloat, floattoindex,      */
/*             chromtointarray_uneq_flds, getfield_uneq_flds,       */
/*             putfield_uneq_flds, map_chromosome_to_std               */
/*             map_chromosome_to_std_uneq_flds,                        */
/*             CombineFileName, Lock, Unlock, MakeLockable, and KillTime. */
/* (routines using other patterns added, others modified to use local maps*/
/* with diffeent field sizes, 5/96, EDG */
/*----------------------------------------------------------------------------*/

#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <stdio.h>
#include "external.h"

#ifndef SEEK_SET
#define SEEK_SET        0       /* Set file pointer to "offset" */
#endif

#ifndef SEEK_CUR
#define SEEK_CUR        1       /* Set file pointer to current plus "offset" */
#endif

#ifndef SEEK_END
#define SEEK_END        2       /* Set file pointer to EOF plus "offset" */
#endif

static int  variabledelay = 0;

void
copyright(parallel)
int     parallel;
{
  int     i, iskip;
  int     ll = 60;
#ifdef BUILDDATE
  FILE *f;
  char tbuf[80];
#endif
  
  iskip = (LINELENGTH - ll) / 2;
  skip(outfp, 1);
  repchar(outfp, " ", iskip);
  repchar(outfp, "-", ll);
  skip(outfp, 1);
  repchar(outfp, " ", iskip);
  fprintf(outfp, 
	  "|       GALOPPS -- \"Genetic ALgorithm Optimized for        |\n");
  repchar(outfp, " ", iskip);
  fprintf(outfp, 
	  "|           Portability and Parallelism\" System            |\n");

  repchar(outfp, " ", iskip);
  if (parallel)
    fprintf(outfp, 
	    "|        Manypops -- Multiple Subpopulation Version        |\n");
  else
    fprintf(outfp, 
	    "|          Onepop -- Single Population Version             |\n");
  repchar(outfp, " ", iskip);

  fprintf(outfp, 
	  "|       (c) 1993, 1994, 1995, 1996   Board of Trustees,    |\n");
  repchar(outfp, " ", iskip);
  fprintf(outfp, 
	  "|               Michigan State University                  |\n");
  repchar(outfp, " ", iskip);
  fprintf(outfp, 
	  "|                    Erik D. Goodman                       |\n");
  repchar(outfp, " ", iskip);
  fprintf(outfp, 
	  "|   Genetic Algorithms Research and Applications Group     |\n");
  repchar(outfp, " ", iskip);
  fprintf(outfp, 
	  "|        Michigan State University, East Lansing           |\n");
  repchar(outfp, " ", iskip);
  fprintf(outfp, 
	  "|                     Release 3.2                          |\n");

  repchar(outfp, " ", iskip);
  fprintf(outfp, 
	  "|  GALOPPS is free software, and is distributed under the  |\n");
  repchar(outfp, " ", iskip);
  fprintf(outfp, 
	  "|     provisions of the GNU Public License.  For more      |\n");
  repchar(outfp, " ", iskip);
  fprintf(outfp, 
	  "|   information, see the file 'license' in subdirectory    |\n");
  repchar(outfp, " ", iskip);
  fprintf(outfp, 
	  "|    'docs'.  THERE IS ABSOLUTELY NO WARRANTY, EXPRESS     |\n");
  repchar(outfp, " ", iskip);
  fprintf(outfp, 
	  "|    OR IMPLIED, ASSOCIATED WITH THIS SOFTWARE.  IT IS     |\n");
  repchar(outfp, " ", iskip);
  fprintf(outfp, 
	  "|          FREE, AND ALL RISK IN USING IT IS YOURS.        |\n");

#ifdef BUILDDATE
  if ((f = fopen("build_date", "r"))) {
    int len, side;
    for (len = 0; (tbuf[len] = getc(f)) != EOF; len++);
    fclose(f);
    tbuf[--len] = '\0';
    side = (ll - len)/2 -1;
    repchar(outfp, " ", iskip);
    fprintf(outfp, "|");
    repchar(outfp, " ", side);
    fprintf(outfp, tbuf);
    repchar(outfp, " ", (ll - len - side - 2));
    fprintf(outfp, "|\n");
  }
#endif

  repchar(outfp, " ", iskip);
  repchar(outfp, "-", ll);
  skip(outfp, 2);
}


void
repchar(outfp, ch, repcount)
/* Repeatedly write a character to stdout */
FILE   *outfp;
char   *ch;
int     repcount;
{
    int     j;

    for (j = 1; j <= repcount; j++)
	fprintf(outfp, "%s", ch);
}

void
skip(outfp, skipcount)
/* Skip skipcount lines */
FILE   *outfp;
int     skipcount;
{
    int     j;

    for (j = 1; j <= skipcount; j++)
	fprintf(outfp, "\n");
}

int
ithruj2int(i, j, from)
/* interpret bits i thru j of a individual as an integer      */
/* j MUST BE greater than or equal to i AND j-i < UINTSIZE-1  */
/* from is a chromosome, represented as an array of unsigneds */
/* ***CAUTION*** This routine numbers bits from 1 up, NOT from */
/* 0 up, although it uses the entire word to pack the information */
int     i, j;
unsigned int    *from;
{
    unsigned mask, temp;
    int     bound_flag;
    int     iisin, jisin;
    int     i1, j1, out;

    if (j < i) {
	fprintf(stderr, "Error in ithruj2int: j < i\n");
	exit(-1);
    }
    if (j - i + 1 > UINTSIZE) {
	fprintf(stderr, "Error in ithruj2int: j-i+1 > UINTSIZE\n");
	exit(-1);
    }

    iisin = (i - 1) / UINTSIZE;
    jisin = (j - 1) / UINTSIZE;

    i1 = i - (iisin * UINTSIZE);
    j1 = j - (jisin * UINTSIZE);

    /* check if bits fall across a word boundary */
    if (iisin == jisin)
	bound_flag = 0;
    else
	bound_flag = 1;

    if (bound_flag == 0) {
	mask = 1;
	mask = (mask << (j1 - i1 + 1)) - 1;
	mask = mask << (i1 - 1);
	out = (from[iisin] & mask) >> (i1 - 1);
	return (out);
    }
    else {
	mask = 1;
	mask = (mask << j1) - 1;
	temp = from[jisin] & mask;
	mask = 1;
	mask = (mask << (UINTSIZE - i1 + 1)) - 1;
	mask = mask << (i1 - 1);
	out = ((from[iisin] & mask) >> (i1 - 1)) | temp << (UINTSIZE - i1 + 1);
	return (out);
    }
}


void
int2ithruj(i, j, valuearg, chrom)
/* Store a positive int value (no more than fieldlength bits) into chrom     */
/* in bits i through j, BUT ***CAUTION***:  This routine uses start and stop */
/* numbering from 1 - lchrom, ***NOT*** 0 - lchrom-1.  Use caution in        */
/* calling.  It stores 16 or 32 bits per word, as usual, however.            */

int     i, j, valuearg;
unsigned *chrom;
{

    unsigned mask;
    int     iisin, jisin;
    int     i1, j1, value;

    value = valuearg;   
    if (j < i) {
	fprintf(stderr, "Error in ithruj2int: j < i\n");
	exit(-1);
    }
    if (j - i + 1 > UINTSIZE) {
	fprintf(stderr, "Error in ithruj2int: j-i+1 > UINTSIZE\n");
	exit(-1);
    }
    /* find out what word or words will write into */
    iisin = (i - 1) / UINTSIZE;
    jisin = (j - 1) / UINTSIZE;

    i1 = i - (iisin * UINTSIZE);
    j1 = j - (jisin * UINTSIZE);

    if (iisin == jisin) {
	/* whole field in one unsigned word, prepare mask for use */
	mask = 1;
	mask = (mask << (j1 - i1 + 1)) - 1;
	mask = mask << (i1 - 1);
	/* use mask to clear field, 'or' shifted value into chrom */
	chrom[iisin] &= (~mask);
	chrom[iisin] |= (value << (i1 - 1));
    }
    else {
	/* copy part of field into jisin, rest into iisin */
	mask = 1;
	mask = (mask << j1) - 1;
	chrom[jisin] &= (~mask);
	chrom[jisin] |= (value >> (UINTSIZE - i1 + 1));
	mask = 1;
	mask = (((mask << (UINTSIZE - i1 + 1)) - 1) << (i1 - 1));
	chrom[iisin] &= (~mask);
	chrom[iisin] |= (value << (i1 - 1));
    }
}


int
getfield(n, nbroffields, chrom)
/* Fetches integer value from field n (numbered 0 - nbroffields-1) of a       */
/* chromosome.  (Works for binary, too.)  Modified 4/96 by EDG such that      */
/* each field may be of a different cardinality (and length), as specified in */
/* array field_sizes[], a global.   */
/* Last bit of field n (in [0, numfields-1]) is bit position field_ends[n],   */
/* (numbered 1,...,lchrom, as in Goldberg's SGA). */

int     n, nbroffields;
unsigned *chrom;

{
    int     start;

    if (n >= nbroffields) {
	fprintf(stderr, " Error in call to getfield, n (= %d) >= nbroffields (= %d)",
		n, nbroffields);
	exit(-1);
    }
    if (n>0)
      start = field_ends[n-1] + 1;
    else
      start = 1;
    return ithruj2int(start, field_ends[n], chrom);
}


int
getfield_uneq_flds(n, nbroffields, chrom, ends_of_fields)
/* New 5/96, EDG, to allow using a different inversion pattern (e.g., one */
/* saved with a best individual) to decode an individual (i.e., one NOT */
/* using the CURRENT map_to_std as used by getfield(). */
/* NOTE:  DO NOT USE THIS UNLESS CURRENT MAP (and field_ends) DOES NOT */
/* MATCH THE LAYOUT OF THE INDIVIDUAL YOU ARE DECODING. */
/* Fetches integer value from field n (numbered 0 - nbroffields-1) of a       */
/* chromosome.  */
/* Last bit of field n (in [0, numfields-1]) is bit position field_ends[n],   */
/* (numbered 1,...,lchrom, as in Goldberg's SGA). */

int     n, nbroffields;
int     *ends_of_fields;
unsigned *chrom;

{
    int     start;

    if (n >= nbroffields) {
	fprintf(stderr, " Error in call to getfield, n (= %d) >= nbroffields (= %d)",
		n, nbroffields);
	exit(-1);
    }
    if (n>0)
      start = ends_of_fields[n-1] + 1;
    else
      start = 1;
    return ithruj2int(start, ends_of_fields[n], chrom);
}


void
putfield(n, nbroffields, value, chrom)
/* Puts integer value into field n (numbered 0 - nbroffields-1) of a          */
/* chromosome.  (Works for binary, too.)  Each field may be of a different    */
/* cardinality (and length), as specified in array field_sizes[], a global.   */
/* Last bit of field n (in [0, numfields-1]) is bit position field_ends[n],   */
/* (numbered 1,...,lchrom, as in Goldberg's SGA). */

int     n, nbroffields, value;
unsigned *chrom;
{
    int     start;

    if (n >= nbroffields) {
	fprintf(stderr, " Error in call to putfield, n (= %d) >= nbroffields (= %d)",
		n, nbroffields);
	exit(-1);
    }
    if (n>0)
      start = field_ends[n-1] + 1;
    else
      start = 1;
    int2ithruj(start, field_ends[n], value, chrom);
}


void
putfield_uneq_flds(n, nbroffields, value, chrom, ends_of_fields)
/* Puts integer value into field n (numbered 0 - nbroffields-1) of a          */
/* chromosome, WHEN THAT CHROMOSOME DOES NOT NECESSARILY SHARE THE SAME */
/* map_to_std (i.e., has been permuted).  Parameter "ends_of_fields" is */
/* the array of field_ends which goes with the chrom passed in.  (Works */
/* for binary, too.)  DO NOT use this if chrom has the current pop's    */
/* inversion pattern, but rather for things like saved "bests" which may */
/* have been computed with a different pattern and preserved.  Each field */
/* may be of a different    */
/* cardinality (and length), as specified in array ends_of_fields[], passed */
/* in.  Last bit of field n (in [0, numfields-1]) is bit position         */
/* ends_of_fields[n], (numbered 1,...,lchrom, as in Goldberg's SGA). */

int     n, nbroffields, value;
int     *ends_of_fields;
unsigned *chrom;
{
    int     start;

    if (n >= nbroffields) {
	fprintf(stderr, " Error in call to putfield, n (= %d) >= nbroffields (= %d)",
		n, nbroffields);
	exit(-1);
    }
    if (n>0)
      start = ends_of_fields[n-1] + 1;
    else
      start = 1;
    int2ithruj(start, ends_of_fields[n], value, chrom);
}


long
powerof2(expon)
/* Returns 2**expon.                                                        */
int     expon;
{
    int     j;
    long    tmp;

    tmp = 1L;
    if (expon >= 0) {
	for (j = 0; j < expon; j++)
	    tmp *= 2;
	return tmp;
    }
    else
	fprintf(stderr, " powerof2 is only for int returns (no neg. args).");
    exit(-1);
}


int
ceilinglogbase2(longarg)
/* calculates ceiling(log base 2 of long argument),positive arguments      */
/* only (sign bit 0).                                                      */
long    longarg;
{
    int     counter = 1;
    long    tmp;

    if (longarg <= 0) {
	fprintf(stderr, "\n*** Call to ceilinglogbase2 with argument <= 0.\n");
	exit(-1);
    }
    tmp = longarg;
    while ((tmp >>= 1) > 1)
	counter++;
    if (tmp << counter != longarg)
	counter++;
    return counter;
}

int
floorlogbase2(integerarg)
/* calculates the floor (log base 2 of integer argument), positive */
/* arguments only (sign bit 0).                                    */
int integerarg;
{
    int     counter = 1;

    if (integerarg <= 0) {
	fprintf(stderr, "\n*** Call to floorlogbase2 with argument <= 0.\n");
	exit(-1);
    }
    if (integerarg == 1) return 0;
    while ((integerarg >>= 1) > 1)
	counter++;
    return counter;

}


void
chromtointarray(intarray, chrom)
/*                  Inverse function to intarraytochrom();                   */
/* Loads intarray with the values from chromosome chrom.                     */
/* Updated for variable-length fields by EDG, 5/96 */
int    *intarray;
unsigned *chrom;
{
    int     i, start;

    start = 1;
    for (i = 0; i < numfields; i++) {
	intarray[i] = ithruj2int(start, field_ends[i], chrom);
	start = field_ends[i] + 1;
    }
}


void
chromtointarray_uneq_flds(intarray, chrom, ends_of_fields)
/* Specialized inverse function to intarraytochrom(); BUT for use when the */                  
/* current map_to_std (and, hence, field_ends) does NOT MATCH the chrom */
/* being decoded (for example, for stored best individuals).  Added 5/96 */
/* by EDG. */
/* Loads intarray with the values from chromosome chrom, for any type of */
/* problem.                                         */
/* Updated for variable-length fields by EDG, 5/96 */
int    *intarray;
int    *ends_of_fields;
unsigned *chrom;
{
    int     i, start;

    start = 1;
    for (i = 0; i < numfields; i++) {
	intarray[i] = ithruj2int(start, ends_of_fields[i], chrom);
	start = ends_of_fields[i] + 1;
    }
}


void
intarraytochrom(intarray, chrom)
/*           Inverse function to chromtointarray();                           */
/* Puts numfields (global) values from intarray into chromosome chrom.        */
/* Updated for variable-length fields by EDG, 5/96 */
int    *intarray;
unsigned *chrom;
{
    int     i, start;

    start = 1;
    for (i = 0; i < numfields; i++) {
	int2ithruj(start, field_ends[i], intarray[i]
		   , chrom);
	start = field_ends[i] + 1;
    }
}


int
hamming(critter1, critter2)
/* Given two chromosomes, calculates Hamming distance between them   */

struct individual *critter1, *critter2;
{
    int     j, k, stop, distance;
    unsigned matches, mask = 1;
    unsigned chrom1, chrom2;

    distance = 0;

    for (k = 0; k < chromsize; k++) {
	chrom1 = critter1->chrom[k];
	chrom2 = critter2->chrom[k];
	matches = chrom1 ^ chrom2; /* 1's wherever chrom bits don't match */
	/* now count the ones to count matches */
	if (k == (chromsize - 1))
	    stop = lchrom - (k * UINTSIZE);
	else
	    stop = UINTSIZE;

	for (j = 0; j < stop; j++) {
	    if (matches & mask)
		distance++;
	    matches = matches >> 1;
	}
    }
    return distance;
}


float
indextofloat(longint, maxlongint, lower, upper)
/* scales a long int in range [0,maxlongint] into a float, linearly, in     */
/* range lower to upper (except if maxlongint ==0, returns 0., warns)       */
long    longint, maxlongint;
float   lower, upper;
{

    if (maxlongint != 0)
	return lower + (upper - lower) * (longint / (float) maxlongint);
    else {
	fprintf(stderr, "\n*** WARNING, BAD maxlongint in call to indextofloat ***\n");
	return 0.;
    }
}

long
floattoindex(floatvalue, lower, upper, maxlongint)
/* maps a float in range [lower, upper] to a long in range [0,maxlongint],  */
/* linearly  (for speed, no safety checks).  Truncates rather than rounding */
long    maxlongint;
float   floatvalue, lower, upper;
{
    long    tmp;

    if (floatvalue <= upper && floatvalue >= lower && upper != lower) {
	tmp = (floatvalue * 1.0000000001 / (upper - lower))
	      * (double) maxlongint;
	return tmp;
    }
    else {
	fprintf(stderr, " Error in call to floattoindex:  bad argument.");
	exit(-1);
    }
}


void
CombineFileName(Buffer, prefix, whichpop, ext)
char   *Buffer;
char   *prefix;
int    whichpop;
char   *ext;
{
    char    tmp[3];

    sprintf(tmp, "%02d", whichpop);
    tmp[2] = '\0';
    strncpy(Buffer, prefix, 6);
    Buffer[6] = '\0';
    strcat(Buffer, tmp);
    strcat(Buffer, ext);
}


/* Provide file locking by creating a companion file with a single flag */
/* int in it, for which 0 means file unlocked and another value>0 tells */
/* the first subpopulation number plus one of the processor which locked it. */
/* Companion filename is same as filename to be locked, except first letter */
/* is replaced by a 'z'. */
BOOL
Lock(filename, num_tries_before_fail)
char *filename;
int  num_tries_before_fail;
{
  FILE *lock_fp;
  char buffer[13], str[80];
  int i, j, lock_flag, chck0, chck1;
  int debug_locking = 0;

  /* Defeat file locking if all subpops using these data files are */
  /* running as a SINGLE process, since it can't cause any conflict.  */
  if (single_process)
    return TRUE;

  /* create pseudonym for lock file name from given file name. */
  /* if filename is abcdefgh.ijk, call the lock file zbcdefg.hij */
  strcpy(buffer, filename);
  buffer[0] = 'z';

  if ((lock_fp = fopen(buffer, "rb"))) 
    fclose(lock_fp); /* just testing */
  else {
    if (debug_locking)
      fprintf(stderr, "File %s does not yet exist.  Cannot lock it.\n", buffer);
    fprintf(outfp, "File %s does not yet exist.  Process %d cannot lock it.\n",
	    buffer, startpopnum);
    return FALSE;
  }
  
  for (i = 0; i < num_tries_before_fail; i++) {
    if ((lock_fp = fopen(buffer, "r+b"))) {
      if (!(fread(&lock_flag, sizeof(int), 1, lock_fp))) {
	if (debug_locking) {
	  if (debug_locking == 1) {
/*          chck0 = ferror(lock_fp);
	    chck1 = feof(lock_fp);
	    fprintf(stderr, "\nRead of lock file failed, listing of %s:",
		    buffer);
	    fprintf(stderr, "\n");
	    sprintf(str, "ls -l %s", buffer);
	    system(str);
*/
	    fprintf(stderr, "\nRead of lock file %s failed... ", buffer);
	  }
	  debug_locking++;
	}
	fclose(lock_fp);
      }

      else if (lock_flag && lock_flag != startpopnum + 1) {
	fclose(lock_fp);
	if (debug_locking) {
	  if (debug_locking == 1)
	    fprintf(stderr, "\n%s locked by process %d... ",
		    filename, lock_flag);
	  debug_locking++;
	}
	KillTime();             /* Was locked, allow more time than usual */
      }
      
      else {
	/* lock the file */
	rewind(lock_fp);
	lock_flag = startpopnum + 1; /* avoid 0 */
	if (!(fwrite(&lock_flag, sizeof(int), 1, lock_fp))) {
	  if (debug_locking) {
	    debug_locking++;
	    fprintf(stderr,"\nCouldn't write lock_flag... ");
	  }
	  fclose(lock_fp);
	}
	
	else {
	  /* Check that the lock was written and NOT overwritten */
	  /* by another processor.  For safety, repeat long enough */
	  /* to avoid a case where another processor also read a 0 */
	  /* but this one wrote and read its flag, and the other one*/
	  /* SUBSEQUENTLY wrote and read its own flag, so BOTH think*/
	  /* they have locked the file.  To prevent it, re-read long*/
	  /* enough so second processor should have written before  */
	  /* this processor reads for the last time (of course, may */
	  /* be system-dependent.  (Really, just killing time       */
	  /* between write and final read.)                         */
	  for (j = 0; j < 10; j++) {
	    fclose(lock_fp);
	    if ((lock_fp = fopen(buffer, "rb"))) {
	      if (!(fread(&lock_flag, sizeof(int), 1, lock_fp))
		  || (lock_flag != startpopnum+1)) {
		if (debug_locking) {
		  if (debug_locking == 1)
		    fprintf(stderr,"\nLock_flag read not = written... ");
		  debug_locking++;
		}
		fclose(lock_fp);
	      }

	      else {
		fclose(lock_fp); /* successfully locked it */
		if (1 < debug_locking) {
		  fprintf(stderr, "lock of %s completed after %d tries.", 
			  filename, debug_locking-1);
		  fflush(stderr);
		}
		return TRUE;
	      }
	    }
	  } /* end for loop */

	}
      }
    } /* end if ((lock_fp = fopen())) */

    else {
      if (debug_locking) {
	if (debug_locking == 1)
	  fprintf(stderr, "\nFile open failed... ");
	  debug_locking++;
	}
      }
    
    KillTime();         /* cycle for another lock attempt */
  } /* end for loop */

  /* No successful lock */
  if (debug_locking) {
    fprintf(stderr, "Unable to lock %s after %d attempts", filename, 
	    debug_locking);
    fflush(stderr);
  }

  return FALSE;
}


/* Inverse function to Lock(); Unlocks a file by writing a 0 as the */
/* contents of its companion file.  Companion file name has first letter 'z', */
/* rest of name is the same as filename to be locked.  Should NEVER be called */
/* unless you currently have successfully locked the file. */
BOOL
Unlock(filename)
     char *filename;
{
  char buffer[13];
  FILE *lock_fp;
  int  int_zero, i;

  if (single_process)
    return TRUE;

  strcpy(buffer, filename);
  buffer[0] = 'z';

  if ((lock_fp = fopen(buffer, "r+b")) == NULL) {
    fprintf(stderr, "\n*** Error opening lockfile when locked.");
    return FALSE;
  }

  int_zero = 0;
  if (fwrite(&int_zero, sizeof(int), 1, lock_fp) != 1) {
    fprintf(stderr, "\n*** Error zeroing lockfile.");
    fclose(lock_fp);
    return FALSE;
  } 

  fclose(lock_fp);
  return TRUE;
}


BOOL
MakeLockable(filename)
char *filename;
/* Create a companion file for the given file name, and make it UNLOCKED */
{

    FILE *lock_fp;
    char buffer[13];
    int i, int_zero, lock_flag;

    if(single_process)
	return TRUE;

    /* create pseudonym for lock file name from given file name. */
    /* if filename is abcdefgh.ijk, call the lock file zbcdefg.hij */
    strcpy(buffer, filename);
    buffer[0] = 'z';

    if((lock_fp = fopen(buffer, "wb")) == NULL) {
	fprintf(stderr, "Problem making %s lockable.", filename);
	exit (-1);
    }
    /* Opened okay, make it say UNLOCKED */
    int_zero = 0;
    if(fwrite(&int_zero, sizeof(int), 1, lock_fp) != 1) {
	fprintf(stderr, "\n** Problem making %s initially unlocked.", filename);
	exit(-1);
    }
    fprintf(outfp, "\nMade a lock file for file %s.\n", filename);
    fclose(lock_fp);
    return TRUE;
}


BOOL
map_chromosome_to_std(chromosome_in, map, chromosome_out)
/* This function uses map and the field_ends for this representation     */
/* (implicit in getfield()) plus the static std_field_ends, to rearrange */
/* the fields on chromosome_in to create chromosome_out, which represents*/ 
/* exactly the same solution, but with different linkage among the fields*/ 
/* (fields re-ordered).  Rewritten for different field sizes 5/96, EDG.  */
unsigned *chromosome_in, *chromosome_out;
int *map;
{
    int i;

    for (i=0;i<numfields;i++) {
	putfield_uneq_flds(map[i]
		 , numfields
		 , getfield(i, numfields, chromosome_in)
		 , chromosome_out, std_field_ends);
    }
    return TRUE;
}

BOOL
map_chromosome_to_std_uneq_flds(chromosome_in, map, chromosome_out
				   , ends_of_fields)
/* This function uses map and the field_ends supplied in the call (and not */
/* necessarily the current ones) plus the static std_field_ends, to rearrange */
/* the fields on chromosome_in to create chromosome_out, which represents*/ 
/* exactly the same solution, but with different linkage among the fields*/ 
/* (fields re-ordered).  Rewritten for different field sizes 5/96, EDG.  */
unsigned *chromosome_in, *chromosome_out;
int *map, *ends_of_fields;
{
    int i;

    for (i=0;i<numfields;i++) {
	putfield_uneq_flds(map[i]
	    , numfields
	    , getfield_uneq_flds(i, numfields, chromosome_in, ends_of_fields)
	    , chromosome_out, std_field_ends);
    }
    return TRUE;
}



/* CAUTION:  Use of definition below may slow your system... alter as needed. */
#ifndef CLOCKS_PER_SEC
#define CLOCKS_PER_SEC 10
#endif

void
KillTime()
/* Function to loop for a random amount of time.  Adjust it if needed to fit  */
/* with your hardware.  Note that the granularity of the actual ticks on your */
/* system may strongly affect the amount of time consumed by KillTime().      */
{
    int delayinmilliseconds;
    double  tstartms, tstopsec, tstarttics;
    double  timesec;

    /* desired delay, different among processes, in ms */
    delayinmilliseconds = 20 + 2 * ((popno + variabledelay++) % 30);
/*  fprintf(stderr, "\nclock() = %ld",clock()); */
    tstarttics = clock();
/*  fprintf(stderr, "\nTstarttics = %f",tstarttics); */
    /* NOTE:  IF, ON SOME OLDER COMPILERS, CLOCKS_PER_SEC IS NOT DEFINED ON */
    /* YOUR SYSTEM, JUST REPLACE IT WITH THE NUMBER OF CLOCK "TICKS" PER    */
    /* SECOND ON YOUR SYSTEM. */
    tstartms = ((double)tstarttics)/CLOCKS_PER_SEC * 1000.;
    tstopsec = (tstartms + delayinmilliseconds)/1000.;
    while ((timesec = clock()/CLOCKS_PER_SEC) < tstopsec) {
/*      fprintf(stderr, "\nDelay, tstart, timesec, tstop = %d, %f, %f, %f"
	     , delayinmilliseconds, tstartms, timesec, tstopsec); */
    }

    return;
}


